@@ -12,6 +12,12 @@ import (
 	"github.com/vmihailenco/msgpack/v4/codes"
 )
 
+const (
+	bytesAllocLimit = 1e6 // 1mb
+	sliceAllocLimit = 1e4
+	maxMapSize      = 1e6
+)
+
 type bufReader interface {
 	io.Reader
 	io.ByteScanner
@@ -30,6 +36,7 @@ func Unmarshal(data []byte, v interface{}) error {
 	return NewDecoder(bytes.NewReader(data)).Decode(v)
 }
 
+// A Decoder reads and decodes MessagePack values from an input stream.
 type Decoder struct {
 	r   io.Reader
 	s   io.ByteScanner
@@ -68,11 +75,17 @@ func (d *Decoder) UseDecodeInterfaceLoose(flag bool) *Decoder {
 
 // UseJSONTag causes the Decoder to use json struct tag as fallback option
 // if there is no msgpack tag.
-func (d *Decoder) UseJSONTag(v bool) *Decoder {
-	d.useJSONTag = v
+func (d *Decoder) UseJSONTag(flag bool) *Decoder {
+	d.useJSONTag = flag
 	return d
 }
 
+// Buffered returns a reader of the data remaining in the Decoder's buffer.
+// The reader is valid until the next call to Decode.
+func (d *Decoder) Buffered() io.Reader {
+	return d.r
+}
+
 func (d *Decoder) Reset(r io.Reader) error {
 	d.resetReader(r)
 	return nil
@@ -477,27 +490,26 @@ func (d *Decoder) readFull(b []byte) error {
 		return err
 	}
 	if d.rec != nil {
+		//TODO: read directly into d.rec?
 		d.rec = append(d.rec, b...)
 	}
 	return nil
 }
 
 func (d *Decoder) readN(n int) ([]byte, error) {
-	buf, err := readN(d.r, d.buf, n)
+	var err error
+	d.buf, err = readN(d.r, d.buf, n)
 	if err != nil {
 		return nil, err
 	}
-	d.buf = buf
 	if d.rec != nil {
 		//TODO: read directly into d.rec?
-		d.rec = append(d.rec, buf...)
+		d.rec = append(d.rec, d.buf...)
 	}
-	return buf, nil
+	return d.buf, nil
 }
 
 func readN(r io.Reader, b []byte, n int) ([]byte, error) {
-	const bytesAllocLimit = 1024 * 1024 // 1mb
-
 	if b == nil {
 		if n == 0 {
 			return make([]byte, 0), nil
@@ -529,7 +541,7 @@ func readN(r io.Reader, b []byte, n int) ([]byte, error) {
 
 		_, err := io.ReadFull(r, b[pos:])
 		if err != nil {
-			return nil, err
+			return b, err
 		}
 
 		if len(b) == n {
