@@ -1,6 +1,8 @@
 package msgpack
 
 import (
+	"encoding"
+	"log"
 	"reflect"
 	"sync"
 
@@ -9,17 +11,30 @@ import (
 
 var errorType = reflect.TypeOf((*error)(nil)).Elem()
 
-var customEncoderType = reflect.TypeOf((*CustomEncoder)(nil)).Elem()
-var customDecoderType = reflect.TypeOf((*CustomDecoder)(nil)).Elem()
+var (
+	customEncoderType = reflect.TypeOf((*CustomEncoder)(nil)).Elem()
+	customDecoderType = reflect.TypeOf((*CustomDecoder)(nil)).Elem()
+)
 
-var marshalerType = reflect.TypeOf((*Marshaler)(nil)).Elem()
-var unmarshalerType = reflect.TypeOf((*Unmarshaler)(nil)).Elem()
+var (
+	marshalerType   = reflect.TypeOf((*Marshaler)(nil)).Elem()
+	unmarshalerType = reflect.TypeOf((*Unmarshaler)(nil)).Elem()
+)
 
-type encoderFunc = func(*Encoder, reflect.Value) error
-type decoderFunc = func(*Decoder, reflect.Value) error
+var (
+	binaryMarshalerType   = reflect.TypeOf((*encoding.BinaryMarshaler)(nil)).Elem()
+	binaryUnmarshalerType = reflect.TypeOf((*encoding.BinaryUnmarshaler)(nil)).Elem()
+)
 
-var typeEncMap sync.Map
-var typeDecMap sync.Map
+type (
+	encoderFunc = func(*Encoder, reflect.Value) error
+	decoderFunc = func(*Decoder, reflect.Value) error
+)
+
+var (
+	typeEncMap sync.Map
+	typeDecMap sync.Map
+)
 
 // Register registers encoder and decoder functions for a value.
 // This is low level API and in most cases you should prefer implementing
@@ -95,22 +110,28 @@ func (f *field) DecodeValue(d *Decoder, strct reflect.Value) error {
 //------------------------------------------------------------------------------
 
 type fields struct {
-	Table   map[string]*field
+	Type    reflect.Type
+	Map     map[string]*field
 	List    []*field
 	AsArray bool
 
 	hasOmitEmpty bool
 }
 
-func newFields(numField int) *fields {
+func newFields(typ reflect.Type) *fields {
 	return &fields{
-		Table: make(map[string]*field, numField),
-		List:  make([]*field, 0, numField),
+		Type: typ,
+		Map:  make(map[string]*field, typ.NumField()),
+		List: make([]*field, 0, typ.NumField()),
 	}
 }
 
 func (fs *fields) Add(field *field) {
-	fs.Table[field.name] = field
+	if _, ok := fs.Map[field.name]; ok {
+		log.Printf("msgpack: %s already has field=%s", fs.Type, field.name)
+	}
+
+	fs.Map[field.name] = field
 	fs.List = append(fs.List, field)
 	if field.omitEmpty {
 		fs.hasOmitEmpty = true
@@ -134,11 +155,10 @@ func (fs *fields) OmitEmpty(strct reflect.Value) []*field {
 }
 
 func getFields(typ reflect.Type, useJSONTag bool) *fields {
-	numField := typ.NumField()
-	fs := newFields(numField)
+	fs := newFields(typ)
 
 	var omitEmpty bool
-	for i := 0; i < numField; i++ {
+	for i := 0; i < typ.NumField(); i++ {
 		f := typ.Field(i)
 
 		tagStr := f.Tag.Get("msgpack")
@@ -181,10 +201,14 @@ func getFields(typ reflect.Type, useJSONTag bool) *fields {
 			if inline {
 				inlineFields(fs, f.Type, field, useJSONTag)
 			} else {
-				inline = autoinlineFields(fs, f.Type, field, useJSONTag)
+				inline = shouldInline(fs, f.Type, field, useJSONTag)
 			}
+
 			if inline {
-				fs.Table[field.name] = field
+				if _, ok := fs.Map[field.name]; ok {
+					log.Printf("msgpack: %s already has field=%s", fs.Type, field.name)
+				}
+				fs.Map[field.name] = field
 				continue
 			}
 		}
@@ -206,7 +230,7 @@ func init() {
 func inlineFields(fs *fields, typ reflect.Type, f *field, useJSONTag bool) {
 	inlinedFields := getFields(typ, useJSONTag).List
 	for _, field := range inlinedFields {
-		if _, ok := fs.Table[field.name]; ok {
+		if _, ok := fs.Map[field.name]; ok {
 			// Don't inline shadowed fields.
 			continue
 		}
@@ -215,7 +239,7 @@ func inlineFields(fs *fields, typ reflect.Type, f *field, useJSONTag bool) {
 	}
 }
 
-func autoinlineFields(fs *fields, typ reflect.Type, f *field, useJSONTag bool) bool {
+func shouldInline(fs *fields, typ reflect.Type, f *field, useJSONTag bool) bool {
 	var encoder encoderFunc
 	var decoder decoderFunc
 
@@ -242,7 +266,7 @@ func autoinlineFields(fs *fields, typ reflect.Type, f *field, useJSONTag bool) b
 
 	inlinedFields := getFields(typ, useJSONTag).List
 	for _, field := range inlinedFields {
-		if _, ok := fs.Table[field.name]; ok {
+		if _, ok := fs.Map[field.name]; ok {
 			// Don't auto inline if there are shadowed fields.
 			return false
 		}
